/* Copyright (C) 2009 Patrick Schmidt, Stefan Frenz
 * 
 * This file is part of SJC, the Small Java Compiler written by Stefan Frenz.
 * 
 * SJC is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * SJC is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with SJC. If not, see <http://www.gnu.org/licenses/>.
 */

package relations;

import compbase.AccVar;
import compbase.Mthd;
import compbase.Unit;

/**
 * RelationManager: keeps track of relations during compilation
 * 
 * @author P. Schmidt, S. Frenz
 * @version 091022 cleanup of interface, added filtering to add-methods
 *  version 091009 initial version
 */

public class RelationManager {
  
  public static final String ROOT_OBJECT="Object";
  
  public RelationElement index=null;
  
  public void removeRelation(Unit u) {
    RelationElement loop, inner;
    //take care if it's the first unit in the list
    while (index!=null && index.unit==u) index=index.next;
    loop=index;
    while (loop!=null) {
      //otherwise check if somewhere in the list
      while (loop.next!=null && loop.next.unit==u) loop.next=loop.next.next;
      //check referencedBy-list
      while (loop.referencedBy!=null && loop.referencedBy.unit==u)
        loop.referencedBy=loop.referencedBy.next;
      inner=loop.referencedBy;
      while (inner!=null) {
        if (inner.next!=null && inner.next.unit==u) inner.next=inner.next.next;
        else inner=inner.next;
      }
      loop=loop.next;
    }
  }
  
  public boolean isStructuralDependance(RelationElement elem) {
    if (elem==null) return false;
    return elem.mthd==null && elem.var==null;
  }
  
  public void addRelation(Unit from, Unit to) {
    if (from.name.equals(ROOT_OBJECT)) return;
    addRelation(from, null, null, to, null);
  }
  
  public void updateRelation(Unit toReplace, Unit with) {
    RelationElement inner, outer=index;
    while (outer!=null) {
      if (outer.unit==toReplace) outer.unit=with;
      inner=outer.referencedBy;
      while (inner!=null) {
        if (inner.unit==toReplace) inner.unit=with;
        inner=inner.next;
      }
      outer=outer.next;
    }
  }
  
  public void addRelation(Unit usedUnit, Mthd usedMthd, AccVar usedVar, Unit usingUnit, Mthd usingMthd) {
    RelationElement outer=index, inner;
    AccVar varToSet=usedVar!=null && usedVar.name!=null ? usedVar : null;
    Mthd mthdToSet=usingMthd!=null && usingMthd.name!=null ? usingMthd : null;
    while (outer!=null) {
      if (outer.unit==usedUnit && outer.mthd==usedMthd && outer.var==varToSet) {
        inner=outer.referencedBy;
        while (inner!=null) {
          if (inner.unit==usingUnit && inner.mthd==mthdToSet) return;
          inner=inner.next;
        }
        break;
      }
      outer=outer.next;
    }
    if (outer==null) {
      outer=new RelationElement();
      outer.unit=usedUnit;//.name;
      outer.mthd=usedMthd;
      outer.var=varToSet;
      outer.next=index;
      inner=outer.referencedBy=new RelationElement();
      index=outer;    
    }
    else {
      inner=new RelationElement();
      inner.next=outer.referencedBy;
      outer.referencedBy=inner;
    }
    inner.unit=usingUnit;//.name;
    inner.mthd=mthdToSet;
  }  
}
